# Copyright (c) 2023-2025 Feud Developers.
# Distributed under the terms of the MIT License (see the LICENSE file).
# SPDX-License-Identifier: MIT
# This source code is part of the Feud project (https://feud.wiki).

import inspect
import os
import re
import sys

import pytest

import feud
import feud.core
from feud import click
from feud import typing as t

sys.path.append(os.path.dirname(os.path.realpath(__file__)))
from fixtures import module  # noqa: E402

overrides: dict[str, t.Any] = {
    "name": "overridden",
    "help": "Overridden help.",
    "epilog": "Overridden epilog.",
    "config": feud.config(negate_flags=False),
}


@pytest.mark.parametrize("override", [False, True])
def test_get_runner_command(*, override: bool) -> None:
    kwargs: dict[str, t.Any] = {}
    if override:
        kwargs.update(overrides)

    runner: click.Command | feud.Group = feud.core.get_runner(
        module.command,
        warn=False,
        **kwargs,
    )

    assert isinstance(runner, click.Command)
    assert runner == module.command

    opt: click.Option = runner.params[0]
    assert opt.type == click.BOOL
    assert opt.name == "opt"
    assert opt.help == "This is an option."
    assert opt.secondary_opts == ["--no-opt"]

    assert runner.name == "command"
    assert runner.help == runner.__func__.__doc__
    assert runner.epilog is None


@pytest.mark.parametrize("override", [False, True])
def test_get_runner_feud_group(*, override: bool) -> None:
    kwargs: dict[str, t.Any] = {}
    if override:
        kwargs.update(overrides)

    runner: click.Command | feud.Group = feud.core.get_runner(
        module.Group,
        warn=False,
        **kwargs,
    )

    assert inspect.isclass(runner)
    assert issubclass(runner, feud.Group)

    click_kwargs = runner.__feud_click_kwargs__
    assert click_kwargs["name"] == "feud-group"
    assert click_kwargs["help"] == runner.__doc__
    assert "epilog" not in click_kwargs
    assert runner.__feud_config__ == feud.config()

    assert runner.__feud_commands__ == ["func"]
    assert runner.func.name == "func"
    assert runner.func.help == runner.func.__func__.__doc__
    assert runner.func.epilog is None

    opt: click.Option = runner.func.params[0]
    assert opt.type == click.BOOL
    assert opt.name == "opt"
    assert opt.help == "This is an option."
    assert opt.secondary_opts == ["--no-opt"]

    assert runner.subgroups() == [module.Subgroup]


@pytest.mark.parametrize("override", [False, True])
def test_get_runner_click_group(*, override: bool) -> None:
    kwargs: dict[str, t.Any] = {}
    if override:
        kwargs.update(overrides)

    runner: click.Command | feud.Group = feud.core.get_runner(
        module.click_group,
        warn=False,
        **kwargs,
    )

    assert isinstance(runner, click.Group)

    assert runner.name == "click-group"
    assert runner.help == "This is a Click group."
    assert runner.epilog is None

    command: click.Command = runner.commands["func"]
    assert command.name == "func"
    assert command.help == command.__func__.__doc__
    assert command.epilog is None

    opt: click.Option = command.params[0]
    assert opt.type == click.BOOL
    assert opt.name == "opt"
    assert opt.help == "This is an option."
    assert opt.secondary_opts == ["--no-opt"]


@pytest.mark.parametrize("override", [False, True])
def test_get_runner_function(*, override: bool) -> None:
    kwargs: dict[str, t.Any] = {}
    if override:
        kwargs.update(overrides)

    runner: click.Command | feud.Group = feud.core.get_runner(
        module.func1,
        warn=False,
        **kwargs,
    )

    assert isinstance(runner, click.Command)

    opt: click.Option = runner.params[0]
    assert opt.type == click.BOOL
    assert opt.name == "opt"
    assert opt.help == "This is an option."

    if override:
        assert runner.name == "overridden"
        assert runner.help == "Overridden help."
        assert runner.epilog == "Overridden epilog."
        assert opt.secondary_opts == []
    else:
        assert runner.name == "func1"
        assert runner.help == runner.__func__.__doc__
        assert runner.epilog is None
        assert opt.secondary_opts == ["--no-opt"]


@pytest.mark.parametrize("override", [False, True])
def test_get_runner_iterable(*, override: bool) -> None:
    kwargs: dict[str, t.Any] = {}
    if override:
        kwargs.update(overrides)

    runner: click.Command | feud.Group = feud.core.get_runner(
        [module.func1, module.command, module.Group, module.click_group],
        warn=False,
        **kwargs,
    )

    # check top-level group settings

    assert inspect.isclass(runner)
    assert issubclass(runner, feud.Group)
    click_kwargs = runner.__feud_click_kwargs__

    if override:
        assert click_kwargs["name"] == "overridden"
        assert click_kwargs["help"] == "Overridden help."
        assert click_kwargs["epilog"] == "Overridden epilog."
        assert runner.__feud_config__ == overrides["config"]
    else:
        assert click_kwargs["name"] == "__feud_group__"
        assert "help" not in click_kwargs
        assert "epilog" not in click_kwargs
        assert runner.__feud_config__ == feud.config()

    assert runner.__feud_commands__ == ["func1", "command", "click-group"]

    # check undecorated function

    func1: click.Command = runner.func1
    assert func1.name == "func1"
    assert func1.help == func1.__func__.__doc__
    assert func1.epilog is None

    opt: click.Option = func1.params[0]
    if override:
        assert opt.secondary_opts == []
    else:
        assert opt.secondary_opts == ["--no-opt"]

    # check decorated command

    command: click.Command = runner.command
    opt: click.Option = command.params[0]
    assert opt.type == click.BOOL
    assert opt.name == "opt"
    assert opt.help == "This is an option."
    assert opt.secondary_opts == ["--no-opt"]

    assert command.name == "command"
    assert command.help == command.__func__.__doc__
    assert command.epilog is None

    # check click.Group

    group: click.Group = getattr(runner, "click-group")
    assert group.name == "click-group"
    assert group.help == "This is a Click group."
    assert group.epilog is None

    command: click.Command = group.commands["func"]
    assert command.name == "func"
    assert command.help == command.__func__.__doc__
    assert command.epilog is None

    opt: click.Option = command.params[0]
    assert opt.type == click.BOOL
    assert opt.name == "opt"
    assert opt.help == "This is an option."
    assert opt.secondary_opts == ["--no-opt"]

    # check feud.Group

    assert runner.subgroups() == [module.Group]
    group: feud.Group = runner.subgroups()[0]

    click_kwargs = group.__feud_click_kwargs__
    assert click_kwargs["name"] == "feud-group"
    assert click_kwargs["help"] == group.__doc__
    assert "epilog" not in click_kwargs
    assert group.__feud_config__ == feud.config()

    assert group.__feud_commands__ == ["func"]
    assert group.func.name == "func"
    assert group.func.help == group.func.__func__.__doc__
    assert group.func.epilog is None

    opt: click.Option = group.func.params[0]
    assert opt.type == click.BOOL
    assert opt.name == "opt"
    assert opt.help == "This is an option."
    assert opt.secondary_opts == ["--no-opt"]

    assert group.subgroups() == [module.Subgroup]


def test_get_runner_iterable_nested() -> None:
    msg = (
        "Groups cannot be constructed from dict or iterable "
        "objects nested within iterable objects."
    )

    with pytest.raises(feud.exceptions.CompilationError, match=re.escape(msg)):
        feud.core.get_runner([module.func1, [module.func1]])

    with pytest.raises(feud.exceptions.CompilationError, match=re.escape(msg)):
        feud.core.get_runner([module.func1, {"func": module.func1}])


@pytest.mark.parametrize("override", [False, True])
def test_get_runner_dict(*, override: bool) -> None:
    kwargs: dict[str, t.Any] = {}
    if override:
        kwargs.update(overrides)

    runner: click.Command | feud.Group = feud.core.get_runner(
        {
            "test-func": module.func1,
            "test-command": module.command,
            "test-feud-group": module.Group,
            "test-click-group": module.click_group,
        },
        warn=False,
        **kwargs,
    )

    # check top-level group settings

    assert inspect.isclass(runner)
    assert issubclass(runner, feud.Group)
    click_kwargs = runner.__feud_click_kwargs__

    if override:
        assert click_kwargs["name"] == "overridden"
        assert click_kwargs["help"] == "Overridden help."
        assert click_kwargs["epilog"] == "Overridden epilog."
        assert runner.__feud_config__ == overrides["config"]
    else:
        assert click_kwargs["name"] == "__feud_group__"
        assert "help" not in click_kwargs
        assert "epilog" not in click_kwargs
        assert runner.__feud_config__ == feud.config()

    assert runner.__feud_commands__ == [
        "test-func",
        "test-command",
        "test-click-group",
    ]

    # check undecorated function

    func: click.Command = getattr(runner, "test-func")
    assert func.name == "test-func"
    assert func.help == func.__func__.__doc__
    assert func.epilog is None

    opt: click.Option = func.params[0]
    if override:
        assert opt.secondary_opts == []
    else:
        assert opt.secondary_opts == ["--no-opt"]

    # check decorated command

    command: click.Command = getattr(runner, "test-command")
    opt: click.Option = command.params[0]
    assert opt.type == click.BOOL
    assert opt.name == "opt"
    assert opt.help == "This is an option."
    assert opt.secondary_opts == ["--no-opt"]

    assert command.name == "test-command"
    assert command.help == command.__func__.__doc__
    assert command.epilog is None

    # check click.Group

    group: click.Group = getattr(runner, "test-click-group")
    assert group.name == "test-click-group"
    assert group.help == "This is a Click group."
    assert group.epilog is None

    command: click.Command = group.commands["func"]
    assert command.name == "func"
    assert command.help == command.__func__.__doc__
    assert command.epilog is None

    opt: click.Option = command.params[0]
    assert opt.type == click.BOOL
    assert opt.name == "opt"
    assert opt.help == "This is an option."
    assert opt.secondary_opts == ["--no-opt"]

    # check feud.Group

    group: feud.Group = runner.subgroups()[0]

    click_kwargs = group.__feud_click_kwargs__
    assert click_kwargs["name"] == "test-feud-group"
    assert click_kwargs["help"] == group.__doc__
    assert "epilog" not in click_kwargs
    assert group.__feud_config__ == feud.config()

    assert group.__feud_commands__ == ["func"]
    assert group.func.name == "func"
    assert group.func.help == group.func.__func__.__doc__
    assert group.func.epilog is None

    opt: click.Option = group.func.params[0]
    assert opt.type == click.BOOL
    assert opt.name == "opt"
    assert opt.help == "This is an option."
    assert opt.secondary_opts == ["--no-opt"]

    assert group.subgroups() == [module.Subgroup]


@pytest.mark.parametrize("override", [False, True])
def test_get_runner_dict_nested(*, override: bool) -> None:
    kwargs: dict[str, t.Any] = {}
    if override:
        kwargs.update(overrides)

    runner: click.Command | feud.Group = feud.core.get_runner(
        {"a": {"b": module.func1, "c": {"d": module.func1}}},
        warn=False,
        **kwargs,
    )

    # check top-level group settings

    assert inspect.isclass(runner)
    assert issubclass(runner, feud.Group)
    click_kwargs = runner.__feud_click_kwargs__

    if override:
        assert click_kwargs["name"] == "overridden"
        assert click_kwargs["help"] == "Overridden help."
        assert click_kwargs["epilog"] == "Overridden epilog."
        assert runner.__feud_config__ == overrides["config"]
    else:
        assert click_kwargs["name"] == "__feud_group__"
        assert "help" not in click_kwargs
        assert "epilog" not in click_kwargs
        assert runner.__feud_config__ == feud.config()

    assert runner.__feud_commands__ == []

    # check subgroup

    subgroups: list[type[feud.Group]] = runner.subgroups()
    assert len(subgroups) == 1
    subgroup: type[feud.Group] = subgroups[0]

    click_kwargs = subgroup.__feud_click_kwargs__
    assert click_kwargs["name"] == "a"
    assert "help" not in click_kwargs
    assert "epilog" not in click_kwargs
    assert subgroup.__feud_config__ == runner.__feud_config__

    # check subgroup commands

    assert subgroup.__feud_commands__ == ["b"]

    command: click.Command = subgroup.b
    assert command.name == "b"
    assert command.help == command.__func__.__doc__
    assert command.epilog is None

    opt: click.Option = command.params[0]
    assert opt.type == click.BOOL
    assert opt.name == "opt"
    assert opt.help == "This is an option."
    if override:
        assert opt.secondary_opts == []
    else:
        assert opt.secondary_opts == ["--no-opt"]

    # check subsubgroup

    subgroups: list[type[feud.Group]] = subgroup.subgroups()
    assert len(subgroups) == 1
    subgroup: type[feud.Group] = subgroups[0]

    click_kwargs = subgroup.__feud_click_kwargs__
    assert click_kwargs["name"] == "c"
    assert "help" not in click_kwargs
    assert "epilog" not in click_kwargs
    assert subgroup.__feud_config__ == runner.__feud_config__

    # check subsubgroup commands

    assert subgroup.__feud_commands__ == ["d"]

    command: click.Command = subgroup.d
    assert command.name == "d"
    assert command.help == command.__func__.__doc__
    assert command.epilog is None

    opt: click.Option = command.params[0]
    assert opt.type == click.BOOL
    assert opt.name == "opt"

    assert opt.help == "This is an option."
    if override:
        assert opt.secondary_opts == []
    else:
        assert opt.secondary_opts == ["--no-opt"]


@pytest.mark.parametrize("override", [False, True])
def test_get_runner_module(*, override: bool) -> None:
    kwargs: dict[str, t.Any] = {}
    if override:
        kwargs.update(overrides)

    runner: click.Command | feud.Group = feud.core.get_runner(
        module, warn=False, **kwargs
    )

    # check top-level group settings

    assert inspect.isclass(runner)
    assert issubclass(runner, feud.Group)
    click_kwargs = runner.__feud_click_kwargs__

    if override:
        assert click_kwargs["name"] == "overridden"
        assert click_kwargs["help"] == "Overridden help."
        assert click_kwargs["epilog"] == "Overridden epilog."
        assert runner.__feud_config__ == overrides["config"]
    else:
        assert click_kwargs["name"] == "module"
        assert click_kwargs["help"] == module.__doc__
        assert "epilog" not in click_kwargs
        assert runner.__feud_config__ == feud.config()

    assert runner.__feud_commands__ == ["func1", "command", "click-group"]

    # check undecorated function

    func1: click.Command = runner.func1
    assert func1.name == "func1"
    assert func1.help == func1.__func__.__doc__
    assert func1.epilog is None

    opt: click.Option = func1.params[0]
    if override:
        assert opt.secondary_opts == []
    else:
        assert opt.secondary_opts == ["--no-opt"]

    # check decorated command

    command: click.Command = runner.command
    opt: click.Option = command.params[0]
    assert opt.type == click.BOOL
    assert opt.name == "opt"
    assert opt.help == "This is an option."
    assert opt.secondary_opts == ["--no-opt"]

    assert command.name == "command"
    assert command.help == command.__func__.__doc__
    assert command.epilog is None

    # check click.Group

    group: click.Group = getattr(runner, "click-group")
    assert group.name == "click-group"
    assert group.help == "This is a Click group."
    assert group.epilog is None

    command: click.Command = group.commands["func"]
    assert command.name == "func"
    assert command.help == command.__func__.__doc__
    assert command.epilog is None

    opt: click.Option = command.params[0]
    assert opt.type == click.BOOL
    assert opt.name == "opt"
    assert opt.help == "This is an option."
    assert opt.secondary_opts == ["--no-opt"]

    # check feud.Group

    assert runner.subgroups() == [module.Group]
    group: feud.Group = runner.subgroups()[0]

    click_kwargs = group.__feud_click_kwargs__
    assert click_kwargs["name"] == "feud-group"
    assert click_kwargs["help"] == group.__doc__
    assert "epilog" not in click_kwargs
    assert group.__feud_config__ == feud.config()

    assert group.__feud_commands__ == ["func"]
    assert group.func.name == "func"
    assert group.func.help == group.func.__func__.__doc__
    assert group.func.epilog is None

    opt: click.Option = group.func.params[0]
    assert opt.type == click.BOOL
    assert opt.name == "opt"
    assert opt.help == "This is an option."
    assert opt.secondary_opts == ["--no-opt"]

    assert group.subgroups() == [module.Subgroup]


@pytest.mark.parametrize("override", [False, True])
def test_get_runner_dict_module(*, override: bool) -> None:
    kwargs: dict[str, t.Any] = {}
    if override:
        kwargs.update(overrides)

    runner: click.Command | feud.Group = feud.core.get_runner(
        {"test-module": module},
        warn=False,
        **kwargs,
    )

    # check top-level group settings

    assert inspect.isclass(runner)
    assert issubclass(runner, feud.Group)
    click_kwargs = runner.__feud_click_kwargs__

    if override:
        assert click_kwargs["name"] == "overridden"
        assert click_kwargs["help"] == "Overridden help."
        assert click_kwargs["epilog"] == "Overridden epilog."
        assert runner.__feud_config__ == overrides["config"]
    else:
        assert click_kwargs["name"] == "__feud_group__"
        assert "help" not in click_kwargs
        assert "epilog" not in click_kwargs
        assert runner.__feud_config__ == feud.config()

    assert runner.__feud_commands__ == []

    # check module

    group: feud.Group = runner.subgroups()[0]

    click_kwargs = group.__feud_click_kwargs__
    assert click_kwargs["name"] == "test-module"
    assert click_kwargs["help"] == module.__doc__
    assert "epilog" not in click_kwargs
    assert group.__feud_commands__ == ["func1", "command", "click-group"]
    if override:
        assert group.__feud_config__ == overrides["config"]
    else:
        assert group.__feud_config__ == feud.config()

    # check undecorated function

    func1: click.Command = group.func1
    assert func1.name == "func1"
    assert func1.help == func1.__func__.__doc__
    assert func1.epilog is None

    opt: click.Option = func1.params[0]
    if override:
        assert opt.secondary_opts == []
    else:
        assert opt.secondary_opts == ["--no-opt"]

    # check decorated command

    command: click.Command = group.command
    opt: click.Option = command.params[0]
    assert opt.type == click.BOOL
    assert opt.name == "opt"
    assert opt.help == "This is an option."
    assert opt.secondary_opts == ["--no-opt"]

    assert command.name == "command"
    assert command.help == command.__func__.__doc__
    assert command.epilog is None

    # check click.Group

    subgroup: click.Group = getattr(group, "click-group")
    assert subgroup.name == "click-group"
    assert subgroup.help == "This is a Click group."
    assert subgroup.epilog is None

    command: click.Command = subgroup.commands["func"]
    assert command.name == "func"
    assert command.help == command.__func__.__doc__
    assert command.epilog is None

    opt: click.Option = command.params[0]
    assert opt.type == click.BOOL
    assert opt.name == "opt"
    assert opt.help == "This is an option."
    assert opt.secondary_opts == ["--no-opt"]


@pytest.mark.parametrize("override", [False, True])
def test_get_runner_dict_iterable(*, override: bool) -> None:
    kwargs: dict[str, t.Any] = {}
    if override:
        kwargs.update(overrides)

    runner: click.Command | feud.Group = feud.core.get_runner(
        {"funcs": [module.func1, module.command]},
        warn=False,
        **kwargs,
    )

    # check top-level group settings

    assert inspect.isclass(runner)
    assert issubclass(runner, feud.Group)
    click_kwargs = runner.__feud_click_kwargs__

    if override:
        assert click_kwargs["name"] == "overridden"
        assert click_kwargs["help"] == "Overridden help."
        assert click_kwargs["epilog"] == "Overridden epilog."
        assert runner.__feud_config__ == overrides["config"]
    else:
        assert click_kwargs["name"] == "__feud_group__"
        assert "help" not in click_kwargs
        assert "epilog" not in click_kwargs
        assert runner.__feud_config__ == feud.config()

    assert runner.__feud_commands__ == []

    # check group settings

    group: feud.Group = runner.subgroups()[0]

    click_kwargs = group.__feud_click_kwargs__
    assert click_kwargs["name"] == "funcs"
    assert "help" not in click_kwargs
    assert "epilog" not in click_kwargs
    assert group.__feud_commands__ == ["func1", "command"]
    if override:
        assert group.__feud_config__ == overrides["config"]
    else:
        assert group.__feud_config__ == feud.config()

    # check undecorated function

    func1: click.Command = group.func1
    assert func1.name == "func1"
    assert func1.help == func1.__func__.__doc__
    assert func1.epilog is None

    opt: click.Option = func1.params[0]
    if override:
        assert opt.secondary_opts == []
    else:
        assert opt.secondary_opts == ["--no-opt"]

    # check decorated command

    command: click.Command = group.command
    opt: click.Option = command.params[0]
    assert opt.type == click.BOOL
    assert opt.name == "opt"
    assert opt.help == "This is an option."
    assert opt.secondary_opts == ["--no-opt"]

    assert command.name == "command"
    assert command.help == command.__func__.__doc__
    assert command.epilog is None



# Copyright (c) 2023-2025 Feud Developers.
# Distributed under the terms of the MIT License (see the LICENSE file).
# SPDX-License-Identifier: MIT
# This source code is part of the Feud project (https://feud.wiki).

from __future__ import annotations

import collections
import typing as t

import pytest

from feud._internal import _types


def annotate(hint: t.Any) -> t.Annotated[t.Any, "annotation"]:
    return t.Annotated[hint, "annotation"]


@pytest.fixture(scope="module")
def helpers(helpers: type) -> type:  # type[Helpers]
    def check_is_collection_type(
        *, annotated: bool, hint: t.Any, expected: tuple[bool, t.Any]
    ) -> None:
        if annotated:
            hint = helpers.annotate(hint)
        assert _types.click.is_collection_type(hint) == expected

    helpers.check_is_collection_type = check_is_collection_type

    return helpers


@pytest.mark.parametrize("annotated", [False, True])
@pytest.mark.parametrize(
    ("hint", "expected"),
    [
        (tuple, (True, None)),
        (t.Tuple, (True, None)),
        (t.Tuple[t.Any], (False, None)),
        (t.Tuple[int, str], (False, None)),
        (t.Tuple[t.Any, ...], (True, t.Any)),
        (t.Tuple[annotate(t.Any), ...], (True, annotate(t.Any))),
    ],
)
def test_tuple(
    helpers: type,
    *,
    annotated: bool,
    hint: t.Any,
    expected: tuple[bool, t.Any],
) -> None:
    helpers.check_is_collection_type(
        annotated=annotated,
        hint=hint,
        expected=expected,
    )


@pytest.mark.parametrize("annotated", [False, True])
@pytest.mark.parametrize(
    ("hint", "expected"),
    [
        (list, (True, None)),
        (t.List, (True, None)),
        (t.List[t.Any], (True, t.Any)),
        (t.List[annotate(t.Any)], (True, annotate(t.Any))),
    ],
)
def test_list(
    helpers: type,
    *,
    annotated: bool,
    hint: t.Any,
    expected: tuple[bool, t.Any],
) -> None:
    helpers.check_is_collection_type(
        annotated=annotated,
        hint=hint,
        expected=expected,
    )


@pytest.mark.parametrize("annotated", [False, True])
@pytest.mark.parametrize(
    ("hint", "expected"),
    [
        (set, (True, None)),
        (t.Set, (True, None)),
        (t.Set[t.Any], (True, t.Any)),
        (t.Set[annotate(t.Any)], (True, annotate(t.Any))),
    ],
)
def test_set(
    helpers: type,
    *,
    annotated: bool,
    hint: t.Any,
    expected: tuple[bool, t.Any],
) -> None:
    helpers.check_is_collection_type(
        annotated=annotated,
        hint=hint,
        expected=expected,
    )


@pytest.mark.parametrize("annotated", [False, True])
@pytest.mark.parametrize(
    ("hint", "expected"),
    [
        (frozenset, (True, None)),
        (t.FrozenSet, (True, None)),
        (t.FrozenSet[t.Any], (True, t.Any)),
        (t.FrozenSet[annotate(t.Any)], (True, annotate(t.Any))),
    ],
)
def test_frozenset(
    helpers: type,
    *,
    annotated: bool,
    hint: t.Any,
    expected: tuple[bool, t.Any],
) -> None:
    helpers.check_is_collection_type(
        annotated=annotated,
        hint=hint,
        expected=expected,
    )


@pytest.mark.parametrize("annotated", [False, True])
@pytest.mark.parametrize(
    ("hint", "expected"),
    [
        (collections.deque, (True, None)),
        (t.Deque, (True, None)),
        (t.Deque[t.Any], (True, t.Any)),
        (t.Deque[annotate(t.Any)], (True, annotate(t.Any))),
    ],
)
def test_deque(
    helpers: type,
    *,
    annotated: bool,
    hint: t.Any,
    expected: tuple[bool, t.Any],
) -> None:
    helpers.check_is_collection_type(
        annotated=annotated,
        hint=hint,
        expected=expected,
    )


@pytest.mark.parametrize("annotated", [False, True])
@pytest.mark.parametrize(
    ("hint", "expected"),
    [
        (t.NamedTuple("Point", x=int, y=str), (False, None)),
        (
            t.NamedTuple("Point", x=annotate(int), y=annotate(str)),
            (False, None),
        ),
    ],
)
def test_namedtuple(
    helpers: type,
    *,
    annotated: bool,
    hint: t.Any,
    expected: tuple[bool, t.Any],
) -> None:
    helpers.check_is_collection_type(
        annotated=annotated,
        hint=hint,
        expected=expected,
    )


@pytest.mark.parametrize("annotated", [False, True])
@pytest.mark.parametrize(
    ("hint", "expected"),
    [(t.Any, (False, None))],
)
def test_other(
    helpers: type,
    *,
    annotated: bool,
    hint: t.Any,
    expected: tuple[bool, t.Any],
) -> None:
    helpers.check_is_collection_type(
        annotated=annotated,
        hint=hint,
        expected=expected,
    )



# Copyright (c) 2023-2025 Feud Developers.
# Distributed under the terms of the MIT License (see the LICENSE file).
# SPDX-License-Identifier: MIT
# This source code is part of the Feud project (https://feud.wiki).

"""Generation of :py:class:`click.Group`s from Python classes,
with automatically generated :py:func:`click.Command`s defined as
methods on the class.
"""

from __future__ import annotations

import copy
import inspect
import types
import typing as t
import warnings
from collections import OrderedDict
from itertools import chain

import pydantic as pyd

import feud.exceptions
from feud import click
from feud._internal import _command, _metaclass
from feud.config import Config
from feud.core.command import build_command_state

__all__ = ["Group", "compile"]


class Group(metaclass=_metaclass.GroupBase):
    """Representation of a command group, compiling into a
    :py:class:`click.Group`.

    Functions defined in the class body represent commands within the group
    and are automatically decorated with :py:func:`.command`
    (if not already decorated and do not begin with an underscore).

    Groups may be registered as subgroups to another parent group.

    Similarly to providing configuration keyword arguments to
    :py:func:`.command` (directly or with a :py:class:`.Config`),
    group-level configuration can be specified when subclassing
    :py:class:`.Group`.

    >>> import feud
    >>> class CLI(feud.Group, show_help_defaults=False, name="my-cli"):
    ...     def func(*, opt: int):
    ...         pass

    Note that Click-level keyword arguments such as ``name``, which are not
    :ref:`Feud configuration parameters <configuration>`, are passed to
    :py:func:`click.group`.

    Feud configuration parameters defined on a group are automatically
    forwarded to the commands within the group, provided that the function in
    the class body is not manually decorated with
    :py:func:`.command`. In the above example, ``func`` is automatically
    wrapped with ``@feud.command(show_help_defaults=False)``.

    .. warning::

        The following function names should **NOT** be used in a group:

        - :py:func:`~deregister`
        - :py:func:`~descendants`
        - :py:func:`~register`
        - :py:func:`~subgroups`

        See :py:func:`.rename` if you wish to define a command with one of the
        above names.
    """

    __feud_config__: t.ClassVar[Config]
    __feud_click_kwargs__: t.ClassVar[dict[str, t.Any]]
    __feud_subgroups__: t.ClassVar[list[type[Group]]]
    __feud_commands__: t.ClassVar[list[str]]

    @staticmethod
    def __new__(
        cls: type[Group], args: list[str] | None = None, /, **kwargs: t.Any
    ) -> t.Any:
        """Compile and run the group.

        Parameters
        ----------
        cls:
            :py:class:`.Group` class reference.

        args:
            Command-line arguments provided to
            :py:class:`click.Command`.

        **kwargs:
            Additional keyword arguments provided to
            :py:class:`click.Command`.

        Returns
        -------
        Output of the called :py:class:`click.Command`.

        Examples
        --------
        >>> import feud
        >>> class CLI(feud.Group):
        ...     def func(*, opt: int) -> int:
        ...         return opt
        >>> CLI(["func", "--opt", "3"], standalone_mode=False)
        3

        See Also
        --------
        .run:
            Run a command or group.
        """
        return cls.__compile__()(args, **kwargs)

    @classmethod
    def __compile__(
        cls: type[Group], *, parent: click.Group | None = None
    ) -> click.Group:
        """Compile the group into a :py:class:`click.Group`.

        .. warning::

            This is an internal function that should not be used directly,
            :py:func:`.compile` should be used instead.

        Parameters
        ----------
        parent:
            Parent :py:class:`click.Group` to attach the compiled
            group to as a subgroup.

        Returns
        -------
        The generated :py:class:`click.Group`.

        Examples
        --------
        >>> import feud, click
        >>> class CLI(feud.Group):
        ...     def func(*, opt: int) -> int:
        ...         return opt
        >>> isinstance(CLI.__compile__(), click.Group)
        True
        """
        # check for circular dependencies
        cls._check_descendants()

        # create the group
        click_group: click.Group = get_group(cls)

        # add commands to the group
        for name in cls.__feud_commands__:
            click_group.add_command(getattr(cls, name))

        # compile all subgroups
        for subgroup in cls.__feud_subgroups__:
            subgroup.__compile__(parent=click_group)

        # add the command group to the parent group if there is one
        if parent:
            parent.add_command(click_group)

        return click_group

    @classmethod
    def subgroups(cls: type[Group]) -> list[type[Group]]:
        """Registered subgroups.

        Returns
        -------
        Registered subgroups.

        Examples
        --------
        >>> import feud
        >>> class A(feud.Group):
        ...     pass
        >>> class B(feud.Group):
        ...     pass
        >>> class C(feud.Group):
        ...     pass
        >>> A.register([B, C])
        >>> A.subgroups()  # doctest: +SKIP
        [<class 'group.B'>, <class 'group.C'>]

        See Also
        --------
        descendants:
            Directed acyclic graph of subgroup descendants.
        """  # noqa: D401
        return list(cls.__feud_subgroups__)

    @classmethod
    def descendants(cls: type[Group]) -> OrderedDict[type[Group], OrderedDict]:
        """Directed acyclic graph of subgroup descendants.

        Returns
        -------
        Subgroup descendants.

        Examples
        --------
        >>> import feud
        >>> class A(feud.Group):
        ...     pass
        >>> class B(feud.Group):
        ...     pass
        >>> class C(feud.Group):
        ...     pass
        >>> A.register(B)
        >>> B.register(C)
        >>> A.descendants()  # doctest: +SKIP
        OrderedDict([
            (
                <class 'group.B'>,
                OrderedDict([
                    (
                        <class 'group.C'>,
                        OrderedDict()
                    )
                ])
            )
        ])

        See Also
        --------
        subgroups:
            Registered subgroups.
        """
        return OrderedDict(
            (group, group.descendants()) for group in cls.__feud_subgroups__
        )

    @classmethod
    def _descendants(cls: type[Group]) -> t.Generator[type[Group]]:
        for group in cls.__feud_subgroups__:
            yield group
            yield from group._descendants()  # noqa: SLF001

    @classmethod
    def _check_descendants(
        cls: type[Group], __target: type[Group] | None = None, /
    ) -> None:
        group: type[Group] = __target or cls
        if cls is __target:
            msg = f"Group {cls.__name__!r} cannot be a subgroup of itself."
            raise feud.RegistrationError(msg)
        if cls in group._descendants():  # noqa: SLF001
            msg = (
                f"Group {cls.__name__!r} is a descendant subgroup of "
                f"{group.__name__!r}, causing a circular dependency."
            )
            raise feud.RegistrationError(msg)

    @classmethod
    def register(
        cls: type[Group],
        sub: type[Group] | list[type[Group]],
        /,
    ) -> None:
        """Register one or more subgroups.

        Parameters
        ----------
        sub:
            The subgroup(s) to register.

        Examples
        --------
        Registering a single subgroup.

        >>> import feud
        >>> class A(feud.Group):
        ...     pass
        >>> class B(feud.Group):
        ...     pass
        >>> A.register(B)
        >>> A.subgroups()
        [<class 'group.B'>]

        Registering multiple subgroups.

        >>> import feud
        >>> class A(feud.Group):
        ...     pass
        >>> class B(feud.Group):
        ...     pass
        >>> class C(feud.Group):
        ...     pass
        >>> A.register([B, C])
        >>> A.subgroups()
        [<class 'group.B'>, <class 'group.C'>]

        See Also
        --------
        deregister:
            Deregister one or more subgroups.
        """
        subgroups: list[type[Group]] = []

        # sub is a list of groups - check each group
        if isinstance(sub, list):
            for group in sub:
                # check if already registered or about to be registered
                if group in cls.__feud_subgroups__ + subgroups:
                    msg = (
                        f"Group {group.__name__!r} is already registered as a "
                        f"subgroup under {cls.__name__!r} and will be ignored."
                    )
                    warnings.warn(msg, RuntimeWarning, stacklevel=1)
                    continue

                # check for circular dependencies
                cls._check_descendants(group)
                subgroups.append(group)

        # sub is a group - check if already registered
        elif sub in cls.__feud_subgroups__:
            msg = (
                f"Group {sub.__name__!r} is already registered as a "
                f"subgroup under {cls.__name__!r} and will be ignored."
            )
            warnings.warn(msg, RuntimeWarning, stacklevel=1)

        # sub is an unregistered group - register it
        else:
            # check for circular dependencies
            cls._check_descendants(sub)
            subgroups.append(sub)

        # update subgroups
        cls.__feud_subgroups__.extend(subgroups)

    @classmethod
    def deregister(
        cls: type[Group],
        sub: type[Group] | list[type[Group]] | None = None,
        /,
    ) -> None:
        """Deregister one or more subgroups.

        Parameters
        ----------
        sub:
            The subgroup(s) to register.

        Examples
        --------
        Deregistering a single subgroup.

        >>> import feud
        >>> class A(feud.Group):
        ...     pass
        >>> class B(feud.Group):
        ...     pass
        >>> A.register(B)
        >>> A.subgroups()
        [<class 'group.B'>]
        >>> A.deregister(B)
        >>> A.subgroups()
        []

        Deregistering multiple subgroups.

        >>> import feud
        >>> class A(feud.Group):
        ...     pass
        >>> class B(feud.Group):
        ...     pass
        >>> class C(feud.Group):
        ...     pass
        >>> A.register([B, C])
        >>> A.subgroups()
        [<class 'group.B'>, <class 'group.C'>]
        >>> A.deregister([B, C])
        >>> A.subgroups()
        []

        See Also
        --------
        register:
            Register one or more subgroups.
        """
        subgroups: list[type[Group]] = []

        if sub:
            # sub is a list of groups - check each group
            if isinstance(sub, list):
                for group in sub:
                    # check if not registered or about to be deregistered
                    if (
                        group not in cls.__feud_subgroups__
                        or group in subgroups
                    ):
                        msg = (
                            f"Group {group.__name__!r} is not a registered "
                            f"subgroup under {cls.__name__!r} and will be "
                            "ignored."
                        )
                        warnings.warn(msg, RuntimeWarning, stacklevel=1)
                        continue
                    subgroups.append(group)
            # sub is a group - check if not registered
            elif sub not in cls.__feud_subgroups__:
                msg = (
                    f"Group {sub.__name__!r} is not a registered subgroup "
                    f"under {cls.__name__!r} and will be ignored."
                )
                warnings.warn(msg, RuntimeWarning, stacklevel=1)
            # sub is a registered group - deregister it
            else:
                subgroups.append(sub)

            # deregister subgroups
            cls.__feud_subgroups__[:] = [
                group
                for group in cls.__feud_subgroups__
                if group not in subgroups
            ]
        else:
            # deregister all subgroups
            cls.__feud_subgroups__ = []

    def __main__() -> None:  # noqa: D105
        pass

    @classmethod
    def from_dict(
        cls: type[Group],
        obj: dict[str, click.Command | type[Group] | t.Callable],
        /,
        **kwargs: t.Any,
    ) -> type[Group]:
        """Create a group from a :py:obj:`dict` of runnable objects.

        Parameters
        ----------
        obj:
            :py:obj:`dict` of runnable function, command or group objects.
        **kwargs:
            Click keywords or Feud configuration to apply:

            - ``name``: :py:obj:`str`
            - ``help``: :py:obj:`str`
            - ``epilog``: :py:obj:`str`
            - ``config``: :py:func:`.config`

        Returns
        -------
        The generated :py:class:`.Group`.
        """
        # split commands and subgroups
        commands: dict[str, click.Command | t.Callable] = obj.copy()
        subgroups: dict[str, type[Group]] = {
            k: commands.pop(k)
            for k, v in obj.items()
            if inspect.isclass(v) and issubclass(v, Group)
        }

        # rename commands (if necessary)
        funcs: list[str] = []
        for name, command in commands.copy().items():
            if isinstance(command, click.Command) and name != command.name:
                # copy command
                commands[name] = copy.copy(command)
                commands[name].name = name
            elif isinstance(command, t.Callable):
                # note commands generated by functions to be renamed later
                if name != command.__name__:
                    funcs.append(name)

        # rename groups
        for name, subgroup in subgroups.copy().items():
            group_name: str = subgroup.__feud_click_kwargs__.get("name")
            if name != group_name:
                # copy (subclass) group
                subgroups[name] = types.new_class(
                    "__feud_group__",
                    bases=(subgroup,),
                    kwds={
                        **subgroup.__feud_click_kwargs__,
                        "name": name,
                    },
                    exec_body=(
                        lambda body: body.update(
                            {"__doc__": subgroup.__doc__},  # noqa: B023
                        )
                    ),
                )

        # create group (only with commands)
        group: type[Group] = types.new_class(
            "__feud_group__",
            bases=(feud.Group,),
            kwds={k: v for k, v in kwargs.items() if v is not None},
            exec_body=(lambda body: body.update(commands)),
        )

        # rename commands generated from functions (if necessary)
        for name in funcs:
            command: click.Command = getattr(group, name)
            command.name = name

        # register subgroups (if any)
        if subgroups:
            group.register(list(subgroups.values()))

        return group

    @classmethod
    def from_iter(
        cls: type[Group],
        obj: t.Iterable[click.Command | type[Group] | t.Callable],
        /,
        **kwargs: t.Any,
    ) -> type[Group]:
        """Create a group from an iterable of runnable objects.

        Parameters
        ----------
        obj:
            :py:obj:`dict` of runnable function, command or group objects.
        **kwargs:
            Click keywords or Feud configuration to apply:

            - ``name``: :py:obj:`str`
            - ``help``: :py:obj:`str`
            - ``epilog``: :py:obj:`str`
            - ``config``: :py:func:`.config`

        Returns
        -------
        The generated :py:class:`.Group`.
        """
        # convert to list
        obj: list[click.Command | type[Group] | t.Callable] = list(obj)

        # split commands and subgroups
        commands: list[click.Command | t.Callable] = obj.copy()
        subgroups: list[type[Group]] = [
            commands.pop(i)
            for i, v in enumerate(obj)
            if inspect.isclass(v) and issubclass(v, Group)
        ]

        # function to get the name of a command or function
        def get_name(o: click.Command | t.Callable) -> str:
            return o.name if isinstance(o, click.Command) else o.__name__

        # group members
        members: dict[str, click.Command | t.Callable] = {
            get_name(cmd): cmd for cmd in commands
        }

        # create group (only with commands)
        group: type[Group] = types.new_class(
            "__feud_group__",
            bases=(feud.Group,),
            kwds={k: v for k, v in kwargs.items() if v is not None},
            exec_body=(lambda body: body.update(members)),
        )

        # register subgroups (if any)
        if subgroups:
            group.register(subgroups)

        return group

    @classmethod
    def from_module(
        cls: type[Group],
        obj: types.ModuleType,
        /,
        **kwargs: t.Any,
    ) -> type[Group]:
        """Create a group from a module of runnable function, command or
        group objects.

        Parameters
        ----------
        obj:
            Module of runnable function, command or group objects.
        **kwargs:
            Click keywords or Feud configuration to apply:

            - ``name``: :py:obj:`str`
            - ``help``: :py:obj:`str`
            - ``epilog``: :py:obj:`str`
            - ``config``: :py:func:`.config`

        Returns
        -------
        The generated :py:class:`.Group`.
        """

        def is_command(item: t.Any) -> bool:
            if inspect.isfunction(item):
                return inspect.getmodule(item) == obj
            return isinstance(item, click.Command)

        def get_descendants(group: type[Group]) -> t.Generator[type[Group]]:
            for name, child in group.descendants().items():
                yield name
                yield from map(get_descendants, child.keys())

        # split function/click.Command/click.Group from feud.Group
        commands: list[t.Callable | click.Command] = []
        groups: list[type[Group]] = []
        for item in obj.__dict__.values():
            if is_command(item):
                commands.append(item)
            elif inspect.isclass(item) and issubclass(item, Group):
                groups.append(item)

        # function to get the name of a command or function
        def get_name(o: click.Command | t.Callable) -> str:
            return o.name if isinstance(o, click.Command) else o.__name__

        # group members
        members: dict[str, click.Command | t.Callable] = {
            get_name(cmd): cmd for cmd in commands
        }

        # add module docstring
        members["__doc__"] = obj.__doc__

        # set group name as module name if none provided
        kwargs["name"] = kwargs.get("name") or obj.__name__.split(".")[-1]

        # discard non-root groups
        non_root: set[type[Group]] = set(
            chain.from_iterable(map(get_descendants, groups))
        )
        subgroups = [group for group in groups if group not in non_root]

        # create group (only with commands)
        group: type[Group] = types.new_class(
            "__feud_group__",
            bases=(feud.Group,),
            kwds={k: v for k, v in kwargs.items() if v is not None},
            exec_body=(lambda body: body.update(members)),
        )

        # register subgroups (if any)
        if subgroups:
            group.register(subgroups)

        return group


@pyd.validate_call(config=pyd.ConfigDict(arbitrary_types_allowed=True))
def compile(group: type[Group], /) -> click.Group:  # noqa: A001
    """Compile a :py:class:`.Group` into a :py:class:`click.Group`.

    Parameters
    ----------
    group:
        Group to compile into a :py:class:`click.Group`.

    Returns
    -------
    The generated :py:class:`click.Group`.

    Examples
    --------
    >>> import feud
    >>> class CLI(feud.Group):
    ...     def func(*, opt: int) -> int:
    ...         return opt
    >>> isinstance(feud.compile(CLI), click.Group)
    True
    """
    return group.__compile__()


def get_group(__cls: type[Group], /) -> click.Group:
    func: callable = __cls.__main__
    if isinstance(func, staticmethod):
        func = func.__func__

    state = _command.CommandState(
        config=__cls.__feud_config__,
        click_kwargs=__cls.__feud_click_kwargs__,
        is_group=True,
        aliases=getattr(func, "__feud_aliases__", {}),
        envs=getattr(func, "__feud_envs__", {}),
        names=getattr(
            func, "__feud_names__", _command.NameDict(command=None, params={})
        ),
        overrides={
            override.name: override
            for override in getattr(func, "__click_params__", [])
        },
    )

    # construct command state from signature
    build_command_state(state, func=func, config=__cls.__feud_config__)

    # generate click.Group and attach original function reference
    command = state.decorate(func)
    command.__func__ = func
    return command



# Copyright (c) 2023-2025 Feud Developers.
# Distributed under the terms of the MIT License (see the LICENSE file).
# SPDX-License-Identifier: MIT
# This source code is part of the Feud project (https://feud.wiki).

from __future__ import annotations

import functools as ft
import inspect
import re
import typing as t

import pydantic as pyd
import pydantic_core as pydc

try:
    import rich_click as click
except ImportError:
    import click


def validate_call(
    func: t.Callable,
    /,
    *,
    name: str,
    param_renames: dict[str, str],
    meta_vars: dict[str, str],
    sensitive_vars: dict[str, bool],
    positional: list[str],
    var_positional: str | None,
    pydantic_kwargs: dict[str, t.Any],
) -> t.Callable:
    @ft.wraps(func)
    def wrapper(*args: t.Any, **kwargs: t.Any) -> t.Callable:
        try:
            # move positional arguments
            for arg in positional:
                pos_arg = kwargs.pop(arg, inspect._empty)  # noqa: SLF001
                if pos_arg is not inspect._empty:  # noqa: SLF001
                    args += (pos_arg,)

            # move *args to positional arguments
            var_pos_args = kwargs.pop(
                var_positional,
                inspect._empty,  # noqa: SLF001
            )
            if var_pos_args is not inspect._empty:  # noqa: SLF001
                args += var_pos_args

            # apply renaming for any options
            inv_mapping = {v: k for k, v in param_renames.items()}
            true_kwargs = {inv_mapping.get(k, k): v for k, v in kwargs.items()}

            # create Pydantic configuration
            config = pyd.ConfigDict(**pydantic_kwargs)

            # validate the function call
            return pyd.validate_call(func, config=config)(*args, **true_kwargs)
        except pyd.ValidationError as e:
            msg = re.sub(
                r"validation error(s?) for (.*)\n",
                rf"validation error\1 for command {name!r}\n",
                str(e),
            )
            for param, meta_var in meta_vars.items():
                msg = re.sub(
                    rf"\n({param})(\.(\d+))?", rf"\n{meta_var} [\3]", msg
                )
                msg = re.sub(r"\s\[\].*\n", "\n", msg)
                msg = re.sub(r"\[type=.*, (input_value=.*)", r"[\1", msg)
                msg = re.sub(r"(.*), input_type=.*\]", r"\1]", msg)
                msg = re.sub(
                    r"\n\s+For further information visit.*(\n?)", r"\1", msg
                )
                if sensitive_vars[param]:
                    msg = re.sub(
                        rf"({meta_var}\s*\n\s.*\[input_value=).*(\])",
                        r"\1hidden\2",
                        msg,
                    )
            raise click.UsageError(msg) from None
        except pydc.SchemaError as e:
            msg = re.sub(
                r'^Error building "call" validator:',
                f"Error building command {name!r}",
                str(e),
            )
            raise click.ClickException(msg) from None

    return wrapper



# Copyright (c) 2023-2025 Feud Developers.
# Distributed under the terms of the MIT License (see the LICENSE file).
# SPDX-License-Identifier: MIT
# This source code is part of the Feud project (https://feud.wiki).

"""Generation of :py:class:`click.Command`s with automatically
defined :py:class:`click.Argument`s and
:py:class:`click.Option`s based on type hints, and help documentation
based on docstrings.
"""

from __future__ import annotations

import inspect
import typing

import docstring_parser
import pydantic as pyd

try:
    import rich_click as click
except ImportError:
    import click

import feud.exceptions
from feud._internal import _command, _types
from feud.config import Config
from feud.typing import custom

__all__ = ["command"]


@pyd.validate_call
def command(
    func: typing.Callable | None = None,
    /,
    *,
    negate_flags: bool | None = None,
    show_help_defaults: bool | None = None,
    show_help_datetime_formats: bool | None = None,
    show_help_envvars: bool | None = None,
    pydantic_kwargs: dict[str, typing.Any] | None = None,
    config: Config | None = None,
    **click_kwargs: typing.Any,
) -> click.Command:
    """Decorate a function and convert it into a
    :py:class:`click.Command` with automatically defined arguments,
    options and help documentation.

    Parameters
    ----------
    func:
        Function used to generate a command.

    negate_flags:
        Whether to automatically add a negated variant for boolean flags.

    show_help_defaults:
        Whether to display default parameter values in command help.

    show_help_datetime_formats:
        Whether to display datetime parameter formats in command help.

    show_help_envvars:
        Whether to display environment variable names in command help.

    pydantic_kwargs:
        Validation settings for
        :py:func:`pydantic.validate_call_decorator.validate_call`.

    config:
        Configuration for the command.

        This argument may be used either in place or in conjunction with the
        other arguments in this function. If a value is provided as both
        an argument to this function, as well as in the provided ``config``,
        the function argument value will take precedence.

    **click_kwargs:
        Keyword arguments to forward :py:func:`click.command`.

    Returns
    -------
    The generated :py:class:`click.Command`.

    Examples
    --------
    >>> import feud
    >>> @feud.command(name="my-command", negate_flags=False)
    ... def f(arg: int, *, opt: int) -> tuple[int, int]:
    ...     return arg, opt
    >>> feud.run(f, ["3", "--opt", "-1"], standalone_mode=False)
    (3, -1)

    See Also
    --------
    .run:
        Run a command or group.

    .Config
        Configuration defaults.
    """

    def decorate(__func: typing.Callable, /) -> typing.Callable:
        # sanitize click kwargs
        _command.sanitize_click_kwargs(click_kwargs, name=__func.__name__)
        # create configuration
        cfg = Config._create(  # noqa: SLF001
            base=config,
            negate_flags=negate_flags,
            show_help_defaults=show_help_defaults,
            show_help_datetime_formats=show_help_datetime_formats,
            show_help_envvars=show_help_envvars,
            pydantic_kwargs=pydantic_kwargs,
        )
        # decorate function
        return get_command(__func, config=cfg, click_kwargs=click_kwargs)

    return decorate(func) if func else decorate


def build_command_state(  # noqa: PLR0915
    state: _command.CommandState, *, func: callable, config: Config
) -> None:
    doc: docstring_parser.Docstring
    if state.is_group:
        doc = docstring_parser.parse(state.click_kwargs.get("help", ""))
    else:
        doc = docstring_parser.parse_from_object(func)

    sig: inspect.Signature = inspect.signature(func)

    for param, spec in sig.parameters.items():
        meta = _command.ParameterSpec()
        meta.hint: type = spec.annotation

        # get renamed parameter if @feud.rename used
        name: str = state.names["params"].get(param, param)

        if _command.pass_context(sig) and param == _command.CONTEXT_PARAM:
            # skip handling for click.Context argument
            state.pass_context = True

        if spec.kind in (spec.POSITIONAL_ONLY, spec.POSITIONAL_OR_KEYWORD):
            # function positional arguments correspond to CLI arguments
            meta.type = _command.ParameterType.ARGUMENT

            # add the argument
            meta.args = [name]

            # special handling for variable-length collections
            is_collection, base_type = _types.click.is_collection_type(
                meta.hint
            )
            if is_collection:
                meta.kwargs["nargs"] = -1
                meta.hint = base_type

            # special handling for feud.typing.custom counting types
            if custom.is_counter(meta.hint):
                msg = (
                    "Counting may only be used in conjunction with "
                    "keyword-only function parameters (command-line "
                    "options), not positional function parameters "
                    "(command-line arguments)."
                )
                raise feud.exceptions.CompilationError(msg)

            # handle option default
            if spec.default is inspect._empty:  # noqa: SLF001
                # specify as required option
                # (if no default provided in function signature)
                meta.kwargs["required"] = True
            else:
                # convert and show default
                # (if default provided in function signature)
                meta.kwargs["default"] = _types.defaults.convert_default(
                    spec.default
                )
        elif spec.kind == spec.KEYWORD_ONLY:
            # function keyword-only arguments correspond to CLI options
            meta.type = _command.ParameterType.OPTION

            # special handling for variable-length collections
            is_collection, base_type = _types.click.is_collection_type(
                meta.hint
            )
            if is_collection:
                meta.kwargs["multiple"] = True
                meta.hint = base_type

            # special handling for feud.typing.custom counting types
            if custom.is_counter(meta.hint):
                meta.kwargs["count"] = True
                meta.kwargs["metavar"] = "COUNT"

            # add the option
            meta.args = [
                _command.get_option(
                    name, hint=meta.hint, negate_flags=config.negate_flags
                )
            ]

            # add aliases - if specified by feud.alias decorator
            for alias in state.aliases.get(param, []):
                meta.args.append(
                    _command.get_alias(
                        alias,
                        hint=meta.hint,
                        negate_flags=config.negate_flags,
                    )
                )

            # add env var - if specified by feud.env decorator
            if env := state.envs.get(param):
                meta.kwargs["envvar"] = env
                meta.kwargs["show_envvar"] = config.show_help_envvars

            # add help - fetch parameter description from docstring
            if doc_param := next(
                (p for p in doc.params if p.arg_name == param), None
            ):
                meta.kwargs["help"] = doc_param.description

            # handle option default
            if spec.default is inspect._empty:  # noqa: SLF001
                # specify as required option
                # (if no default provided in function signature)
                meta.kwargs["required"] = True
            else:
                # convert and show default
                # (if default provided in function signature)
                meta.kwargs["show_default"] = config.show_help_defaults
                meta.kwargs["default"] = _types.defaults.convert_default(
                    spec.default
                )
        elif spec.kind == spec.VAR_POSITIONAL:
            # function positional arguments correspond to CLI arguments
            meta.type = _command.ParameterType.ARGUMENT

            # add the argument
            meta.args = [name]

            # special handling for variable-length collections
            meta.kwargs["nargs"] = -1

            # special handling for feud.typing.custom counting types
            if custom.is_counter(meta.hint):
                msg = (
                    "Counting may only be used in conjunction with "
                    "keyword-only function parameters (command-line "
                    "options), not positional function parameters "
                    "(command-line arguments)."
                )
                raise feud.exceptions.CompilationError(msg)

        # add the parameter
        if meta.type == _command.ParameterType.ARGUMENT:
            state.arguments[param] = meta
        elif meta.type == _command.ParameterType.OPTION:
            state.options[param] = meta


def get_command(
    func: typing.Callable,
    /,
    *,
    config: Config,
    click_kwargs: dict[str, typing.Any],
) -> click.Command:
    if isinstance(func, staticmethod):
        func = func.__func__

    state = _command.CommandState(
        config=config,
        click_kwargs=click_kwargs,
        is_group=False,
        aliases=getattr(func, "__feud_aliases__", {}),
        envs=getattr(func, "__feud_envs__", {}),
        names=getattr(
            func, "__feud_names__", _command.NameDict(command=None, params={})
        ),
        overrides={
            override.name: override
            for override in getattr(func, "__click_params__", [])
        },
    )

    # construct command state from signature
    build_command_state(state, func=func, config=config)

    # generate click.Command and attach original function reference
    command = state.decorate(func)
    command.__func__ = func
    return command



# Copyright (c) 2023-2025 Feud Developers.
# Distributed under the terms of the MIT License (see the LICENSE file).
# SPDX-License-Identifier: MIT
# This source code is part of the Feud project (https://feud.wiki).

"""Override for :py:class:`click.Context`."""

from __future__ import annotations

import typing as t

import click
from pydantic_core import core_schema

__all__ = ["Context"]


class Context(click.Context):
    """Override :py:class:`click.Context` for type hints that skip
    validation.

    Example
    -------
    >>> import feud
    >>> from feud import click
    >>> class CLI(feud.Group):
    ...     def f(*, arg: int) -> int:
    ...         return arg
    ...     def g(ctx: click.Context, *, arg: int) -> int:
    ...         return ctx.forward(CLI.f)
    >>> feud.run(CLI, ["g", "--arg", "3"], standalone_mode=False)
    3
    """

    @classmethod
    def __get_pydantic_core_schema__(
        cls: type[Context], _source: t.Any, _handler: t.Any
    ) -> core_schema.CoreSchema:
        """Override pydantic schema to validate as typing.Any."""
        return core_schema.any_schema()



# Copyright (c) 2023-2025 Feud Developers.
# Distributed under the terms of the MIT License (see the LICENSE file).
# SPDX-License-Identifier: MIT
# This source code is part of the Feud project (https://feud.wiki).

"""Tasks for running coverage checks."""

from invoke.config import Config
from invoke.tasks import task


@task
def install(c: Config) -> None:
    """Install package with core and coverage dependencies."""
    c.run("poetry install --sync --only base,main,cov -E extra-types -E email")



# Copyright (c) 2023-2025 Feud Developers.
# Distributed under the terms of the MIT License (see the LICENSE file).
# SPDX-License-Identifier: MIT
# This source code is part of the Feud project (https://feud.wiki).

"""Configuration for defining commands using :py:func:`.command` or
groups using :py:class:`.Group`.
"""

from __future__ import annotations

import inspect
from typing import Any

import pydantic as pyd

__all__ = ["Config", "config"]


class Config(pyd.BaseModel):
    """Class representing a reusable configuration for
    :py:func:`.command` and :py:class:`.Group` objects.

    .. warning::

        This class should **NOT** be instantiated directly ---
        :py:func:`.config` should be used to create a :py:class:`.Config`
        instead.
    """

    #: Whether to automatically add a negated variant for boolean flags.
    negate_flags: bool = True

    #: Whether to display default parameter values in command help.
    show_help_defaults: bool = True

    #: Whether to display datetime parameter formats in command help.
    show_help_datetime_formats: bool = False

    #: Whether to display environment variable names in command help.
    show_help_envvars: bool = True

    #: Validation settings for
    #: :py:func:`pydantic.validate_call_decorator.validate_call`.
    pydantic_kwargs: dict[str, Any] = {}

    def __init__(self: Config, **kwargs: Any) -> Config:
        caller: str = inspect.currentframe().f_back.f_code.co_name
        if caller != Config._create.__name__:
            msg = (
                "The feud.Config class should not be instantiated directly, "
                "the feud.config function should be used instead."
            )
            raise RuntimeError(msg)
        super().__init__(**kwargs)

    @classmethod
    def _create(
        cls: type[Config], base: Config | None = None, **kwargs: Any
    ) -> Config:
        config_kwargs = base.model_dump(exclude_unset=True) if base else {}
        for field in cls.model_fields:
            value: Any | None = kwargs.get(field)
            if value is not None:
                config_kwargs[field] = value
        return cls(**config_kwargs)


def config(
    *,
    negate_flags: bool | None = None,
    show_help_defaults: bool | None = None,
    show_help_datetime_formats: bool | None = None,
    show_help_envvars: bool | None = None,
    pydantic_kwargs: dict[str, Any] | None = None,
) -> Config:
    """Create a reusable configuration for :py:func:`.command` or
    :py:class:`.Group` objects.

    See :py:class:`.Config` for the underlying configuration class.

    Parameters
    ----------
    negate_flags:
        Whether to automatically add a negated variant for boolean flags.

    show_help_defaults:
        Whether to display default parameter values in command help.

    show_help_datetime_formats:
        Whether to display datetime parameter formats in command help.

    show_help_envvars:
        Whether to display environment variable names in command help.

    pydantic_kwargs:
        Validation settings for
        :py:func:`pydantic.validate_call_decorator.validate_call`.

    Returns
    -------
    The reusable :py:class:`.Config`.

    Examples
    --------
    Providing a configuration to :py:func:`.command`.

    >>> import feud
    >>> config = feud.config(show_help_defaults=False)
    >>> @feud.command(config=config)
    ... def func(*, opt1: int, opt2: bool = True):
    ...     pass
    >>> all(not param.show_default for param in func.params)
    True

    Providing a configuration to :py:class:`.Group`.

    Note that the configuration is internally forwarded to the commands
    defined within the group.

    >>> import feud
    >>> config = feud.config(show_help_defaults=False)
    >>> class CLI(feud.Group, config=config):
    ...     def func(*, opt1: int, opt2: bool = True):
    ...         pass
    >>> all(not param.show_default for param in CLI.func.params)
    True
    """
    return Config._create(  # noqa: SLF001
        negate_flags=negate_flags,
        show_help_defaults=show_help_defaults,
        show_help_datetime_formats=show_help_datetime_formats,
        show_help_envvars=show_help_envvars,
        pydantic_kwargs=pydantic_kwargs,
    )



# Copyright (c) 2023-2025 Feud Developers.
# Distributed under the terms of the MIT License (see the LICENSE file).
# SPDX-License-Identifier: MIT
# This source code is part of the Feud project (https://feud.wiki).

import os
import re
from typing import Callable
from unittest import mock

import pytest

import feud
from feud import click
from feud import typing as t


def assert_help(
    __obj: t.Union[
        feud.Group,
        click.Group,
        click.Command,
        Callable,
    ],
    /,
    *,
    capsys: pytest.CaptureFixture,
    expected: str,
) -> None:
    with pytest.raises(SystemExit):
        feud.run(__obj, ["--help"])
    out, _ = capsys.readouterr()
    assert out.strip() == expected.strip()


@pytest.fixture(scope="module")
def env_command() -> Callable:
    @feud.env(opt1="OPT1", opt2="OPT2", opt3="OPT3")
    def f(
        *, opt1: t.PositiveInt, opt2: bool, opt3: t.NegativeFloat
    ) -> tuple[t.PositiveInt, bool, t.NegativeFloat]:
        """Returns a full path.\f

        Parameters
        ----------
        opt1:
            First option.
        opt2:
            Second option.
        opt3:
            Third option.
        """
        return opt1, opt2, opt3

    return f


def test_valid_format() -> None:
    @feud.command
    @feud.alias(arg1="-a", arg2="-b")
    def f(*, arg1: int, arg2: bool) -> None:
        pass

    assert [param.opts for param in f.params] == [
        ["--arg1", "-a"],
        ["--arg2", "-b"],
    ]


@pytest.mark.parametrize("negate_flags", [True, False])
def test_multiple_aliases(*, negate_flags: bool) -> None:
    @feud.command(negate_flags=negate_flags)
    @feud.alias(arg1=["-a", "-A"])
    def f(*, arg1: bool) -> None:
        pass

    assert f.params[0].opts == ["--arg1", "-a", "-A"]

    secondary_opts = ["--no-arg1", "--no-a", "--no-A"] if negate_flags else []
    assert f.params[0].secondary_opts == secondary_opts


def test_multiple_aliases_non_unique() -> None:
    with pytest.raises(feud.CompilationError):

        @feud.command
        @feud.alias(arg1=["-a", "-a"])
        def f(*, arg1: int, arg2: bool) -> None:
            pass


def test_undefined_alias() -> None:
    with pytest.raises(feud.CompilationError):

        @feud.alias(arg1="-a", arg2="-b")
        def f(*, arg1: int) -> None:
            pass


def test_invalid_format_alias() -> None:
    with pytest.raises(feud.CompilationError):

        @feud.alias(arg1="-a1")
        def f(*, arg1: int) -> None:
            pass


def test_alias_argument() -> None:
    with pytest.raises(feud.CompilationError):

        @feud.alias(arg1="-a")
        def f(arg1: int) -> None:
            pass


def test_undefined_env() -> None:
    with pytest.raises(feud.CompilationError):

        @feud.env(opt1="-a", opt2="-b")
        def f(*, opt1: int) -> None:
            pass


def test_env_help_with_show(
    capsys: pytest.CaptureFixture, env_command: Callable
) -> None:
    assert_help(
        feud.command(show_help_envvars=True)(env_command),
        capsys=capsys,
        expected="""
Usage: pytest [OPTIONS]

  Returns a full path.

Options:
  --opt1 INTEGER RANGE  First option.  [env var: OPT1; x>0; required]
  --opt2 / --no-opt2    Second option.  [env var: OPT2; required]
  --opt3 FLOAT RANGE    Third option.  [env var: OPT3; x<0; required]
  --help                Show this message and exit.
        """,
    )


def test_env_help_without_show(
    capsys: pytest.CaptureFixture, env_command: Callable
) -> None:
    assert_help(
        feud.command(show_help_envvars=False)(env_command),
        capsys=capsys,
        expected="""
Usage: pytest [OPTIONS]

  Returns a full path.

Options:
  --opt1 INTEGER RANGE  First option.  [x>0; required]
  --opt2 / --no-opt2    Second option.  [required]
  --opt3 FLOAT RANGE    Third option.  [x<0; required]
  --help                Show this message and exit.
        """,
    )


def test_env_call_no_env(env_command: Callable) -> None:
    with pytest.raises(click.MissingParameter):
        feud.run(env_command, [], standalone_mode=False)


@mock.patch.dict(os.environ, {"OPT1": "long"}, clear=True)
def test_env_call_hidden() -> None:
    @feud.env(opt1="OPT1")
    def f(*, opt1: t.constr(max_length=3)) -> None:
        pass

    msg = "String should have at most 3 characters [input_value=hidden]"
    with pytest.raises(click.UsageError, match=re.escape(msg)):
        feud.run(f, [], standalone_mode=False)


@mock.patch.dict(
    os.environ, {"OPT1": "1", "OPT2": "true", "OPT3": "-0.1"}, clear=True
)
def test_env_call_with_env(env_command: Callable) -> None:
    assert feud.run(env_command, [], standalone_mode=False) == (1, True, -0.1)


@mock.patch.dict(os.environ, {"OPT": "long"}, clear=True)
def test_override_env_hidden() -> None:
    @click.option("--opt", type=str, envvar="OPT")
    def f(*, opt: t.constr(max_length=3)) -> str:
        return opt

    msg = "String should have at most 3 characters [input_value=hidden]"
    with pytest.raises(click.UsageError, match=re.escape(msg)):
        feud.run(f, [], standalone_mode=False)


def test_rename_command() -> None:
    @feud.command
    @feud.rename("func")
    def f(*, opt: int) -> int:
        return opt

    assert f.name == "func"


def test_rename_params() -> None:
    @feud.rename(arg1="arg-1", arg2="arg-2", opt1="opt-1", opt2="opt-2")
    def f(
        ctx: click.Context, arg1: int, arg2: str, *, opt1: bool, opt2: float
    ) -> None:
        return arg1, arg2, opt1, opt2

    cmd = feud.command(f)

    # check arg1 -> arg-1 rename
    assert cmd.params[0].name == "arg-1"

    # check arg2 -> arg-2 rename
    assert cmd.params[1].name == "arg-2"

    # check opt1 -> opt-1 rename
    # should create options --opt-1/--no-opt-1
    assert cmd.params[2].name == "opt-1"
    assert cmd.params[2].opts == ["--opt-1"]
    assert cmd.params[2].secondary_opts == ["--no-opt-1"]

    # check opt2 -> opt-2 rename
    # should create option --opt-2
    assert cmd.params[3].name == "opt-2"
    assert cmd.params[3].opts == ["--opt-2"]

    # test call
    assert cmd(
        ["2", "test", "--no-opt-1", "--opt-2", "0.2"], standalone_mode=False
    ) == (2, "test", False, 0.2)


def test_rename_command_and_params(capsys: pytest.CaptureFixture) -> None:
    @feud.rename(
        "func", arg1="arg-1", arg2="arg-2", opt1="opt-1", opt2="opt-2"
    )
    def f(
        ctx: click.Context, arg1: int, arg2: str, *, opt1: bool, opt2: float
    ) -> None:
        return arg1, arg2, opt1, opt2

    cmd = feud.command(f)

    # check command name
    assert cmd.name == "func"

    # check arg1 -> arg-1 rename
    assert cmd.params[0].name == "arg-1"

    # check arg2 -> arg-2 rename
    assert cmd.params[1].name == "arg-2"

    # check opt1 -> opt-1 rename
    # should create options --opt-1/--no-opt-1
    assert cmd.params[2].name == "opt-1"
    assert cmd.params[2].opts == ["--opt-1"]
    assert cmd.params[2].secondary_opts == ["--no-opt-1"]

    # check opt2 -> opt-2 rename
    # should create option --opt-2
    assert cmd.params[3].name == "opt-2"
    assert cmd.params[3].opts == ["--opt-2"]

    # check help
    assert_help(
        cmd,
        capsys=capsys,
        expected="""
Usage: pytest [OPTIONS] ARG-1 ARG-2

Options:
  --opt-1 / --no-opt-1  [required]
  --opt-2 FLOAT         [required]
  --help                Show this message and exit.
        """,
    )

    # test call
    assert cmd(
        ["2", "test", "--no-opt-1", "--opt-2", "0.2"], standalone_mode=False
    ) == (2, "test", False, 0.2)


@mock.patch.dict(os.environ, {"OPT1": "1", "OPT2": "true"}, clear=True)
def test_all_decorators(capsys: pytest.CaptureFixture) -> None:
    @feud.rename("cmd", opt1="opt-1", opt2="opt-2", opt3="opt_3")
    @feud.env(opt1="OPT1", opt2="OPT2")
    @feud.alias(opt3="-o")
    def command(
        *, opt1: t.PositiveInt, opt2: bool, opt3: t.NegativeFloat
    ) -> t.Path:
        """Returns a full path.\f

        Parameters
        ----------
        opt1:
            First option.
        opt2:
            Second option.
        opt3:
            Third option.
        """
        return opt1, opt2, opt3

    cmd = feud.command(command)

    # check command name
    assert cmd.name == "cmd"

    # check opt1 -> opt-1 rename
    # should create option --opt-1
    assert cmd.params[0].name == "opt-1"
    assert cmd.params[0].opts == ["--opt-1"]
    assert cmd.params[0].envvar == "OPT1"

    # check opt2 -> opt-2 rename
    # should create option --opt-2
    assert cmd.params[1].name == "opt-2"
    assert cmd.params[1].opts == ["--opt-2"]
    assert cmd.params[1].envvar == "OPT2"

    # check opt3 -> opt_3 rename
    # should create option --opt_3
    assert cmd.params[2].name == "opt_3"
    assert cmd.params[2].opts == ["--opt_3", "-o"]

    # check help
    assert_help(
        cmd,
        capsys=capsys,
        expected="""
Usage: pytest [OPTIONS]

  Returns a full path.

Options:
  --opt-1 INTEGER RANGE    First option.  [env var: OPT1; x>0; required]
  --opt-2 / --no-opt-2     Second option.  [env var: OPT2; required]
  -o, --opt_3 FLOAT RANGE  Third option.  [x<0; required]
  --help                   Show this message and exit.
        """,
    )

    # test call
    assert cmd(["--opt_3", "-1.2"], standalone_mode=False) == (1, True, -1.2)


def test_rename_group(capsys: pytest.CaptureFixture) -> None:
    class Test(feud.Group):
        @staticmethod
        @feud.rename("test-group", opt1="opt-1")
        def __main__(ctx: click.Context, *, opt1: int) -> None:
            ctx.obj = {"opt1": opt1}

        @staticmethod
        @feud.rename("func", opt2="opt_2")
        def f(ctx: click.Context, *, opt2: int) -> int:
            return ctx.obj["opt1"], opt2

    assert Test(
        ["--opt-1", "1", "func", "--opt_2", "2"],
        standalone_mode=False,
    ) == (1, 2)



# Copyright (c) 2023-2025 Feud Developers.
# Distributed under the terms of the MIT License (see the LICENSE file).
# SPDX-License-Identifier: MIT
# This source code is part of the Feud project (https://feud.wiki).

import re

import pytest

import feud
from feud import click
from feud import typing as t


def test_help_docstring(capsys: pytest.CaptureFixture) -> None:
    @feud.command
    def f(arg1: int, *, arg2: bool) -> None:
        """Do something."""

    with pytest.raises(SystemExit):
        f(["--help"])

    out, _ = capsys.readouterr()

    assert (
        out.strip()
        == """
Usage: pytest [OPTIONS] ARG1

  Do something.

Options:
  --arg2 / --no-arg2  [required]
  --help              Show this message and exit.
    """.strip()
    )


def test_help_click_kwarg(capsys: pytest.CaptureFixture) -> None:
    @feud.command(help="Do something.")
    def f(arg1: int, *, arg2: bool) -> None:
        pass

    with pytest.raises(SystemExit):
        f(["--help"])

    out, _ = capsys.readouterr()

    assert (
        out.strip()
        == """
Usage: pytest [OPTIONS] ARG1

  Do something.

Options:
  --arg2 / --no-arg2  [required]
  --help              Show this message and exit.
    """.strip()
    )


def test_param_help_docstring(capsys: pytest.CaptureFixture) -> None:
    @feud.command
    def f(arg1: int, *, arg2: bool) -> None:
        """Do something.\f

        Parameters
        ----------
        arg1:
            Controls something.

        arg2:
            Changes something.
        """

    with pytest.raises(SystemExit):
        f(["--help"])

    out, _ = capsys.readouterr()

    assert (
        out.strip()
        == """
Usage: pytest [OPTIONS] ARG1

  Do something.

Options:
  --arg2 / --no-arg2  Changes something.  [required]
  --help              Show this message and exit.
    """.strip()
    )


def test_param_help_click_override(capsys: pytest.CaptureFixture) -> None:
    @feud.command
    @click.option(
        "--arg2/--no-arg2", type=bool, required=True, help="Changes something."
    )
    def f(arg1: int, *, arg2: bool) -> None:
        """Do something."""

    with pytest.raises(SystemExit):
        f(["--help"])

    out, _ = capsys.readouterr()

    assert (
        out.strip()
        == """
Usage: pytest [OPTIONS] ARG1

  Do something.

Options:
  --arg2 / --no-arg2  Changes something.  [required]
  --help              Show this message and exit.
    """.strip()
    )


def test_no_call() -> None:
    @feud.command
    def f(arg1: int, *, arg2: bool) -> None:
        pass

    assert isinstance(f, click.Command)


def test_empty_call() -> None:
    @feud.command()
    def f(arg1: int, *, arg2: bool) -> None:
        pass

    assert isinstance(f, click.Command)


def test_kwarg_call() -> None:
    @feud.command(config=feud.config())
    def f(arg1: int, *, arg2: bool) -> None:
        pass

    assert isinstance(f, click.Command)


def test_config_propagation() -> None:
    """Check that configuration options propagate to command parameters.

    Only if not overriden.
    """

    @feud.command(show_help_defaults=True)
    def f(*, arg1: bool = True, arg2: int = 1) -> None:
        pass

    assert all(param.show_default is True for param in f.params)

    @feud.command(show_help_defaults=False)
    def f(*, arg1: bool = True, arg2: int = 1) -> None:
        pass

    assert all(not param.show_default for param in f.params)

    @feud.command(show_help_defaults=False)
    @click.option("--arg2", type=int, default=1, show_default=True)
    def f(*, arg1: bool = True, arg2: int = 1) -> None:
        pass

    assert f.params[0].show_default is False
    assert f.params[1].show_default is True


def test_forward() -> None:
    @feud.command
    def g(*, arg1: bool) -> bool:
        return arg1

    @feud.command
    def f(ctx: click.Context, *, arg1: bool) -> bool:
        return ctx.forward(g)

    assert not f(["--no-arg1"], standalone_mode=False)


def test_invoke() -> None:
    @feud.command
    def g(*, arg1: bool) -> bool:
        return arg1

    @feud.command
    def f(ctx: click.Context, *, arg1: bool) -> bool:
        return ctx.invoke(g, arg1=arg1)

    assert not f(["--no-arg1"], standalone_mode=False)


def test_run_undecorated() -> None:
    def f(arg1: int, *, arg2: bool) -> tuple[int, bool]:
        return arg1, arg2

    assert feud.run(f, ["1", "--no-arg2"], standalone_mode=False) == (1, False)


def test_run_decorated() -> None:
    @feud.command
    def f(arg1: int, *, arg2: bool) -> tuple[int, bool]:
        return arg1, arg2

    assert feud.run(f, ["1", "--no-arg2"], standalone_mode=False) == (1, False)


def test_sensitive_input() -> None:
    @feud.command
    @click.option("--password", type=str, hide_input=True)
    def f(*, password: t.constr(min_length=10)) -> None:
        pass

    msg = """1 validation error for command 'f'
--password
  String should have at least 10 characters [input_value=hidden]
    """.strip()
    with pytest.raises(click.UsageError, match=re.escape(msg)):
        feud.run(f, ["--password", "abc"], standalone_mode=False)


@pytest.mark.parametrize("command", [True, False])
def test_run_dict(capsys: pytest.CaptureFixture, *, command: bool) -> None:
    def first(arg1: int, *, opt1: float, opt2: bool = False) -> None:
        """The first command.\f

        Parameters
        ----------
        arg1:
            An argument.
        opt1:
            The first option.
        opt2:
            The second option.
        """

    def second(*, opt: int) -> None:
        """The second command.\f

        Parameters
        ----------
        opt:
            An option.
        """

    if command:
        first = feud.command()(first)
        second = feud.command()(second)

    with pytest.raises(SystemExit):
        feud.run({"1st": first, "2nd": second}, ["--help"])

    out, _ = capsys.readouterr()

    assert (
        out.strip()
        == """
Usage: pytest [OPTIONS] COMMAND [ARGS]...

Options:
  --help  Show this message and exit.

Commands:
  1st  The first command.
  2nd  The second command.
        """.strip()
    )

    with pytest.raises(SystemExit):
        feud.run({"1st": first, "2nd": second}, ["1st", "--help"])

    out, _ = capsys.readouterr()

    assert (
        out.strip()
        == """
Usage: pytest 1st [OPTIONS] ARG1

  The first command.

Options:
  --opt1 FLOAT        The first option.  [required]
  --opt2 / --no-opt2  The second option.  [default: no-opt2]
  --help              Show this message and exit.
        """.strip()
    )


@pytest.mark.parametrize("command", [True, False])
def test_run_iterable(capsys: pytest.CaptureFixture, *, command: bool) -> None:
    def first(arg1: int, *, opt1: float, opt2: bool = False) -> None:
        """The first command.\f

        Parameters
        ----------
        arg1:
            An argument.
        opt1:
            The first option.
        opt2:
            The second option.
        """

    def second(*, opt: int) -> None:
        """The second command.\f

        Parameters
        ----------
        opt:
            An option.
        """

    if command:
        first = feud.command()(first)
        second = feud.command()(second)

    with pytest.raises(SystemExit):
        feud.run((first, second), ["--help"])

    out, _ = capsys.readouterr()

    assert (
        out.strip()
        == """
Usage: pytest [OPTIONS] COMMAND [ARGS]...

Options:
  --help  Show this message and exit.

Commands:
  first   The first command.
  second  The second command.
        """.strip()
    )

    with pytest.raises(SystemExit):
        feud.run((first, second), ["first", "--help"])

    out, _ = capsys.readouterr()

    assert (
        out.strip()
        == """
Usage: pytest first [OPTIONS] ARG1

  The first command.

Options:
  --opt1 FLOAT        The first option.  [required]
  --opt2 / --no-opt2  The second option.  [default: no-opt2]
  --help              Show this message and exit.
        """.strip()
    )


def test_full_signature(
    capsys: pytest.CaptureFixture,
) -> None:
    @feud.command
    def command(
        a: float,
        /,
        b: str,
        *c: t.PositiveInt,
        d: int,
        e: bool = True,
        **f: float,
    ) -> None:
        """Does something.\f

        Parameters
        ----------
        a:
            Test 1.
        b:
            Test 2.
        *c:
            Test 3.
        d:
            Test 4.
        e:
            Test 5.
        **f:
            Test 6.
        """
        return a, b, c, d, e, f

    # check params (**f should be ignored)
    params = command.params
    assert len(params) == 5

    a = command.params[0]
    assert isinstance(a, click.Argument)
    assert a.name == "a"
    assert a.type == click.FLOAT
    assert a.opts == ["a"]
    assert a.secondary_opts == []
    assert a.nargs == 1
    assert a.required
    assert a.default is None

    b = command.params[1]
    assert isinstance(b, click.Argument)
    assert b.name == "b"
    assert b.type == click.STRING
    assert b.opts == ["b"]
    assert b.secondary_opts == []
    assert b.nargs == 1
    assert b.required
    assert b.default is None

    c = command.params[2]
    assert isinstance(c, click.Argument)
    assert c.name == "c"
    assert isinstance(c.type, click.IntRange)
    assert c.type.min == 0
    assert c.type.min_open
    assert c.opts == ["c"]
    assert c.secondary_opts == []
    assert c.nargs == -1
    assert not c.required
    assert c.default is None

    d = command.params[3]
    assert isinstance(d, click.Option)
    assert d.name == "d"
    assert d.help == "Test 4."
    assert d.type == click.INT
    assert d.opts == ["--d"]
    assert d.secondary_opts == []
    assert d.nargs == 1
    assert d.required
    assert d.default is None

    e = command.params[4]
    assert isinstance(e, click.Option)
    assert e.name == "e"
    assert e.help == "Test 5."
    assert e.type == click.BOOL
    assert e.opts == ["--e"]
    assert e.secondary_opts == ["--no-e"]
    assert e.nargs == 1
    assert not e.required
    assert e.default is True


def test_argument_default() -> None:
    @feud.command
    def f(
        a: int,
        /,
        b: float,
        c: bool = True,  # noqa: FBT001, FBT002
    ) -> None:
        pass

    assert len(f.params) == 3

    a = f.params[0]
    assert isinstance(a, click.Argument)
    assert a.name == "a"
    assert a.type == click.INT
    assert a.opts == ["a"]
    assert a.secondary_opts == []
    assert a.nargs == 1
    assert a.required
    assert a.default is None

    b = f.params[1]
    assert isinstance(b, click.Argument)
    assert b.name == "b"
    assert b.type == click.FLOAT
    assert b.opts == ["b"]
    assert b.secondary_opts == []
    assert b.nargs == 1
    assert b.required
    assert b.default is None

    c = f.params[2]
    assert isinstance(c, click.Argument)
    assert c.name == "c"
    assert c.type == click.BOOL
    assert c.opts == ["c"]
    assert c.secondary_opts == []
    assert c.nargs == 1
    assert not c.required
    assert c.default is True



# Copyright (c) 2023-2025 Feud Developers.
# Distributed under the terms of the MIT License (see the LICENSE file).
# SPDX-License-Identifier: MIT
# This source code is part of the Feud project (https://feud.wiki).

"""Officially supported types from the below standard library packages.

- ``collections``
- ``datetime``
- ``decimal``
- ``enum``
- ``pathlib``
- ``uuid``
"""

from collections import deque
from datetime import date, datetime, time, timedelta
from decimal import Decimal
from enum import Enum, IntEnum, StrEnum
from pathlib import Path
from uuid import UUID



# Copyright (c) 2023-2025 Feud Developers.
# Distributed under the terms of the MIT License (see the LICENSE file).
# SPDX-License-Identifier: MIT
# This source code is part of the Feud project (https://feud.wiki).

import pytest

from feud import typing as t
from feud._internal import _types
from feud.config import Config


@pytest.fixture(scope="module")
def config() -> Config:
    return Config._create()  # noqa: SLF001


@pytest.fixture(scope="module")
def helpers(helpers: type) -> type:  # type[Helpers]
    def is_lambda(__obj: t.Any, /) -> bool:
        return (
            callable(__obj)
            and getattr(__obj, "__name__", None) == (lambda: None).__name__
        )

    # register is_lambda helper
    helpers.is_lambda = is_lambda

    def check_get_click_type(
        *,
        config: Config,
        annotated: bool,
        hint: t.Any,
        expected: t.Tuple[bool, t.Any],
    ) -> None:
        if annotated:
            hint = helpers.annotate(hint)

        result = _types.click.get_click_type(hint, config=config)

        if helpers.is_lambda(expected):
            assert expected(result)
        else:
            assert result == expected

    # register check_get_click_type helper
    helpers.check_get_click_type = check_get_click_type

    return helpers



# Copyright (c) 2023-2025 Feud Developers.
# Distributed under the terms of the MIT License (see the LICENSE file).
# SPDX-License-Identifier: MIT
# This source code is part of the Feud project (https://feud.wiki).

"""Package exceptions."""

__all__ = ["CompilationError", "RegistrationError"]


class CompilationError(Exception):
    """An exception indicating an issue that would prevent a
    :py:class:`click.Command` command from being generated as expected.
    """


class RegistrationError(Exception):
    """An exception relating to the registering or deregistering of
    :py:class:`.Group` classes.
    """



# Copyright (c) 2023-2025 Feud Developers.
# Distributed under the terms of the MIT License (see the LICENSE file).
# SPDX-License-Identifier: MIT
# This source code is part of the Feud project (https://feud.wiki).

"""Officially supported types from the ``pydantic-extra-types`` package."""

try:
    from pydantic_extra_types.color import Color
    from pydantic_extra_types.coordinate import (
        Coordinate,
        Latitude,
        Longitude,
    )
    from pydantic_extra_types.country import (
        CountryAlpha2,
        CountryAlpha3,
        CountryNumericCode,
        CountryOfficialName,
        CountryShortName,
    )
    from pydantic_extra_types.mac_address import MacAddress
    from pydantic_extra_types.payment import (
        PaymentCardBrand,
        PaymentCardNumber,
    )
    from pydantic_extra_types.phone_numbers import PhoneNumber
    from pydantic_extra_types.routing_number import ABARoutingNumber
except ImportError:
    pass



# Copyright (c) 2023-2025 Feud Developers.
# Distributed under the terms of the MIT License (see the LICENSE file).
# SPDX-License-Identifier: MIT
# This source code is part of the Feud project (https://feud.wiki).

import pytest

from feud.config import Config


def test_create_no_base_no_kwargs() -> None:
    """No keyword arguments set."""
    config = Config._create()  # noqa: SLF001
    assert config.negate_flags is True
    assert config.show_help_defaults is True
    assert config.show_help_datetime_formats is False
    assert config.show_help_envvars is True
    assert config.pydantic_kwargs == {}


def test_create_no_base_none_kwargs() -> None:
    """Keyword arguments set to None."""
    config = Config._create(  # noqa: SLF001
        negate_flags=None,
        show_help_defaults=None,
        show_help_datetime_formats=None,
        show_help_envvars=None,
        pydantic_kwargs=None,
    )
    assert config.negate_flags is True
    assert config.show_help_defaults is True
    assert config.show_help_datetime_formats is False
    assert config.show_help_envvars is True
    assert config.pydantic_kwargs == {}


def test_create_no_base_with_kwargs_default() -> None:
    """Keyword arguments set to default values."""
    config = Config._create(  # noqa: SLF001
        negate_flags=True,
        show_help_defaults=True,
    )
    assert config.negate_flags is True
    assert config.show_help_defaults is True


def test_create_no_base_with_kwargs_non_default() -> None:
    """Keyword arguments set to non-default values."""
    config = Config._create(  # noqa: SLF001
        negate_flags=False,
        show_help_defaults=False,
    )
    assert config.negate_flags is False
    assert config.show_help_defaults is False


def test_create_base_with_no_base_kwargs_no_override_kwargs() -> None:
    """Base configuration with no base keyword arguments set
    and no override keyword arguments set.
    """
    base = Config._create()  # noqa: SLF001
    config = Config._create(base=base)  # noqa: SLF001
    assert config == base


def test_create_base_with_no_base_kwargs_override_kwargs() -> None:
    """Base configuration with no base keyword arguments set
    and with override keyword arguments set.
    """
    base = Config._create()  # noqa: SLF001
    config = Config._create(  # noqa: SLF001
        base=base,
        negate_flags=False,
        show_help_defaults=False,
    )
    assert config.negate_flags is False
    assert config.show_help_defaults is False


def test_create_base_with_base_kwargs_default_no_override_kwargs() -> None:
    """Base configuration with base keyword arguments set to default values
    and no override keyword arguments set.
    """
    base = Config._create(  # noqa: SLF001
        negate_flags=True,
        show_help_defaults=True,
    )
    config = Config._create(base=base)  # noqa: SLF001
    assert config == base


def test_create_base_with_base_kwargs_default_override_kwargs() -> None:
    """Base configuration with base keyword arguments set to default values
    and override keyword arguments set.
    """
    base = Config._create(  # noqa: SLF001
        negate_flags=True,
        show_help_defaults=True,
    )
    config = Config._create(  # noqa: SLF001
        base=base,
        negate_flags=False,
        show_help_defaults=False,
    )
    assert config.negate_flags is False
    assert config.show_help_defaults is False


def test_create_base_with_base_kwargs_non_default_no_override_kwargs() -> None:
    """Base configuration with base keyword arguments set to non-default values
    and no override keyword arguments set.
    """
    base = Config._create(  # noqa: SLF001
        negate_flags=False,
        show_help_defaults=False,
    )
    config = Config._create(base=base)  # noqa: SLF001
    assert config == base


def test_create_base_with_base_kwargs_non_default_override_kwargs() -> None:
    """Base configuration with base keyword arguments set to non-default values
    and override keyword arguments set.
    """
    base = Config._create(  # noqa: SLF001
        negate_flags=False,
        show_help_defaults=False,
    )
    config = Config._create(  # noqa: SLF001
        base=base,
        negate_flags=False,
        show_help_defaults=False,
    )
    assert config.negate_flags is False
    assert config.show_help_defaults is False


# test instantiate
def test_init_create() -> None:
    """Should not be able to instantiate feud.Config."""
    with pytest.raises(RuntimeError):
        Config()
